use std::fs::File;
use std::io::{ErrorKind, Read};
use std::net::IpAddr;
use std::path::Path;
use serde::Deserialize;
use validator::Validate;
use anyhow::{bail, Context, Result};
use mac_address::MacAddress;

#[derive(Deserialize, Validate)]
pub struct Mqtt {
    #[validate(length(min = 1))]
    pub host: String,
    pub port: u16,
    pub client_id: String,

    #[serde(default)]
    pub credentials: Option<MqttCredentials>,
}

#[derive(Deserialize)]
pub struct MqttCredentials {
    pub user: String,
    pub password: String,
}

#[derive(Deserialize, Validate, Clone)]
pub struct Host {
    #[validate(length(min = 1))]
    #[serde(default)]
    pub id: Option<String>,
    pub display_name: String,
    pub mac_address: MacAddress,

    #[serde(default)]
    pub hass_identifiers: Vec<String>,

    #[validate(range(min = 1))]
    #[serde(default = "default_packet_count")]
    pub packet_count: usize,

    #[serde(default = "default_ms_between_packets")]
    pub ms_between_packets: usize,
}

fn default_packet_count() -> usize {
    6
}

fn default_ms_between_packets() -> usize {
    500
}

#[derive(Deserialize, Validate)]
pub struct Config {
    #[validate]
    pub mqtt: Mqtt,
    pub broadcast_address: IpAddr,

    #[validate]
    #[serde(rename = "host", default)] // more natural in TOML syntax
    pub hosts: Vec<Host>
}

pub fn load(config_file_path: &Path) -> Result<Config> {
    match File::open(config_file_path) {
        Ok(mut file) => {
            log::info!("Reading configuration file: {}", config_file_path.to_string_lossy());

            let mut string_content = String::new();
            file.read_to_string(&mut string_content).context("while reading the configuration file")?;

            let parsed = toml::from_str::<Config>(string_content.as_str()).context("while parsing the configuration file")?;
            parsed.validate().context("while validating the configuration file")?;

            Ok(parsed)
        }

        Err(error) if error.kind() == ErrorKind::NotFound => {
            bail!("The configuration file does not exist: {}", config_file_path.to_string_lossy());
        }

        Err(error) => Err(error).context("while opening the configuration file"),
    }
}