use std::collections::{HashMap, HashSet};
use std::env;
use std::net::IpAddr;
use std::path::{Path, PathBuf};
use std::process::exit;
use std::str::FromStr;
use std::time::Duration;

use anyhow::Result;
use mac_address::MacAddress;
use rumqttc::AsyncClient as MqttClient;
use tokio::net::UdpSocket;
use tokio::time::sleep;

use crate::config::Config;
use crate::mqtt::{MqttMessageHandler, OwnedTopicsService};

mod config;
mod mqtt;

fn initialize_logger() {
    let mut builder = env_logger::builder();
    builder.parse_filters("warn,hass_wol=info");
    builder.parse_default_env();
    builder.init();
}

#[tokio::main]
async fn main() -> Result<()> {
    initialize_logger();

    let args = env::args().collect::<Vec<_>>();

    if args.iter().any(|a| a == "-h" || a == "--help") {
        print_help();
        exit(exitcode::OK);
    }

    if args.iter().any(|a| a == "-v" || a == "--version") {
        println!("hass-wol v{} is licensed under the Blue Oak Model License 1.0.0.", env!("CARGO_PKG_VERSION"));
        println!("See {} for more information.", env!("CARGO_PKG_HOMEPAGE"));
        exit(exitcode::OK);
    }

    if let Some(config_argument) = args.get(1) {
        if let Some(data_directory_argument) = args.get(2) {
            if args.len() == 3 {
                if let Ok(config_file_path) = PathBuf::from_str(config_argument) {
                    if let Ok(data_directory_path) = PathBuf::from_str(data_directory_argument) {
                        let config = config::load(&config_file_path)?;
                        start(config, &data_directory_path).await?;
                        exit(exitcode::OK)
                    } else {
                        eprintln!("Invalid persistent data directory path.");
                    }
                } else {
                    eprintln!("Invalid config file path.");
                }

                exit(exitcode::CONFIG);
            }
        }
    }

    print_help();
    exit(exitcode::USAGE);
}

fn print_help() {
    println!("Syntax: hass-wol <configuration file> <persistent data directory> OR hass-wol [--help] [--version]\n");
    println!("See {} for more information.", env!("CARGO_PKG_HOMEPAGE"));
}

pub struct InitializationContext {
    // Owned topics are topics which a retained message was sent into.
    pub owned_mqtt_topics: HashSet<String>,
    pub message_handler_by_mqtt_topic: HashMap<String, Box<MqttMessageHandler>>,
    pub owned_topics_service: OwnedTopicsService,
    pub mqtt_client: MqttClient,
    pub config: Config,
    pub availability_topic: String,
}

impl InitializationContext {
    fn subscribe_mqtt_topic<F: Fn(&str) -> Result<()> + 'static>(&mut self, topic: impl Into<String>, handler: F) {
        self.message_handler_by_mqtt_topic.insert(topic.into(), Box::new(handler));
    }

    async fn publish_to_owned_mqtt_topic(&mut self, topic: &str, message: impl Into<String>) -> std::result::Result<(), rumqttc::ClientError> {
        let message = message.into();

        self.register_owned_mqtt_topic(topic);
        self.mqtt_client.publish(topic, rumqttc::QoS::AtLeastOnce, true, message).await
    }

    fn register_owned_mqtt_topic(&mut self, topic: &str) {
        self.owned_mqtt_topics.insert(topic.into());
    }
}

async fn start(config: Config, data_directory_path: &Path) -> Result<()> {
    if config.hosts.is_empty() {
        log::warn!("No hosts are defined in the configuration file.");
    }

    let availability_topic = format!("hass-wol/{}/availability", config.mqtt.client_id);
    let (mqtt_client, event_loop) = mqtt::create_client(&config, &availability_topic).await?;

    let mut context = InitializationContext {
        owned_mqtt_topics: HashSet::new(),
        message_handler_by_mqtt_topic: HashMap::new(),
        owned_topics_service: OwnedTopicsService::new(data_directory_path).await?,
        mqtt_client,
        config,
        availability_topic
    };

    let hosts = context.config.hosts.clone();

    for host in hosts {
        log::debug!("Registering host: {}", &host.mac_address);
        let id = host.id.clone().unwrap_or(host.mac_address.to_string().replace(":", ""));
        let global_id = format!("wol_{}", &id);
        let base_topic = format!("homeassistant/button/{}/{}", &context.config.mqtt.client_id, &id);
        let command_topic = format!("{}/trigger", &base_topic);

        context.publish_to_owned_mqtt_topic(
            &format!("{}/config", &base_topic),
            json::stringify(json::object! {
                "availability_topic": context.availability_topic.as_str(),
                "command_topic": command_topic.as_str(),
                "device": {
                    "connections": [["mac", host.mac_address.to_string()]],
                    "identifiers": host.hass_identifiers,
                    "name": host.display_name.as_str()
                },
                "name": "Wake-on-LAN",
                "icon": "mdi:power",
                "object_id": global_id.as_str(),
                "unique_id": global_id.as_str()
            })
        ).await?;

        context.subscribe_mqtt_topic(command_topic, move |text| {
            if text == "PRESS" {
                let broadcast_address = context.config.broadcast_address;
                tokio::spawn(async move {
                    if let Err(error) = wake(
                        broadcast_address,
                        host.mac_address,
                        host.packet_count,
                        Duration::from_millis(host.ms_between_packets as u64)
                    ).await {
                        log::error!("{:#}", error);
                    }
                });
            }

            Ok(())
        })
    }

    mqtt::start_communication(context, event_loop).await?;
    Ok(())
}

async fn wake(broadcast_address: IpAddr, mac_address: MacAddress, packet_count: usize, time_between_packets: Duration) -> Result<()> {
    log::info!("Waking {} with {} packet(s)", mac_address, packet_count);
    let socket = UdpSocket::bind("0.0.0.0:0").await?;
    socket.set_broadcast(true)?;
    let mut data: [u8; 102] = [0xFF; 102];

    for i in 1..=16 {
        data[(i * 6)..((i + 1) * 6)].copy_from_slice(&mac_address.bytes());
    }

    for _ in 0..packet_count {
        socket.send_to(&data, (broadcast_address, 9)).await?;
        sleep(time_between_packets).await;
    }

    Ok(())
}
